from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn
from contextlib import contextmanager
import sys

from deepeval.telemetry import capture_synthesizer_run


@contextmanager
def progress_context(
    description: str, total: int = 9999, transient: bool = True
):
    console = Console(file=sys.stderr)  # Direct output to standard error
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        console=console,  # Use the custom console
        transient=transient,
    ) as progress:
        progress.add_task(description=description, total=total)
        yield


@contextmanager
def synthesizer_progress_context(
    evaluation_model: str,
    embedder: str = None,
    max_generations: str = None,
    use_case: str = "QA",
    _show_indicator: bool = True,
):
    with capture_synthesizer_run(max_generations):
        if embedder is None:
            description = f"✨ 🍰 ✨ You're generating up to {max_generations} goldens using DeepEval's latest Synthesizer (using {evaluation_model}, use case={use_case})! This may take a while..."
        else:
            description = f"✨ 🍰 ✨ You're generating up to {max_generations} goldens using DeepEval's latest Synthesizer (using {evaluation_model} and {embedder}, use case={use_case})! This may take a while..."
        console = Console(file=sys.stderr)  # Direct output to standard error
        if _show_indicator:
            with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                console=console,  # Use the custom console
                transient=True,
            ) as progress:
                progress.add_task(description=description, total=9999)
                yield
        else:
            yield



import pytest
import os
import json
from rich import print
from typing import Optional, Any
from deepeval.constants import PYTEST_RUN_TEST_NAME
from deepeval.test_run import test_run_manager, DeploymentConfigs
from deepeval.utils import get_is_running_deepeval


def pytest_sessionstart(session: pytest.Session):
    is_running_deepeval = get_is_running_deepeval()

    if is_running_deepeval:
        test_run_manager.save_to_disk = True
        try:
            deployment_configs = session.config.getoption("--deployment")
            disable_request = False

            if deployment_configs is None:
                deployment = False
            else:
                deployment = True
                deployment_configs = json.loads(deployment_configs)
                disable_request = deployment_configs.pop(
                    "is_pull_request", False
                )
                deployment_configs = DeploymentConfigs(**deployment_configs)

            test_run_manager.create_test_run(
                deployment=deployment,
                deployment_configs=deployment_configs,
                file_name=session.config.getoption("file_or_dir")[0],
                disable_request=disable_request,
            )
        except:
            test_run_manager.create_test_run()


def pytest_addoption(parser):
    parser.addoption(
        "--deployment",
        action="store",
        default=None,
        help="Set deployment configs",
    )


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_protocol(
    item: pytest.Item, nextitem: Optional[pytest.Item]
) -> Optional[Any]:
    os.environ[PYTEST_RUN_TEST_NAME] = item.nodeid.split("::")[-1]
    return None


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_sessionfinish(session: pytest.Session, exitstatus):
    print("Running teardown with pytest sessionfinish...")

    yield


def pytest_terminal_summary(terminalreporter, exitstatus, config):
    for report in terminalreporter.getreports("skipped"):
        if report.skipped:
            reason = report.longreprtext.split("\n")[-1]
            print(f"Test {report.nodeid} was skipped. Reason: {reason}")



from .base_metric import BaseMetric, BaseConversationalMetric

from .bias.bias import BiasMetric
from .toxicity.toxicity import ToxicityMetric
from .hallucination.hallucination import HallucinationMetric
from .answer_relevancy.answer_relevancy import AnswerRelevancyMetric
from .summarization.summarization import SummarizationMetric
from .g_eval.g_eval import GEval
from .faithfulness.faithfulness import FaithfulnessMetric
from .contextual_recall.contextual_recall import ContextualRecallMetric
from .contextual_relevancy.contextual_relevancy import ContextualRelevancyMetric
from .contextual_precision.contextual_precision import ContextualPrecisionMetric
from .knowledge_retention.knowledge_retention import KnowledgeRetentionMetric

# from .ragas_metric import (
#     RagasMetric,
#     RAGASAnswerRelevancyMetric,
#     RAGASFaithfulnessMetric,
#     RAGASContextualRecallMetric,
#     RAGASContextualRelevancyMetric,
#     RAGASContextualPrecisionMetric,
#     RAGASAnswerRelevancyMetric,
# )



import pytest
from deepeval.metrics import BaseMetric
from deepeval.test_case import LLMTestCase
from deepeval import assert_test
import random
import time


class FakeMetric(BaseMetric):
    def __init__(
        self,
        threshold: float = 0.5,
        include_reason=True,
        model="asd",
        strict_mode=True,
    ):
        self.threshold = threshold
        self.include_reason = include_reason
        self.evaluation_model = model
        self.strict_mode = strict_mode

    def measure(self, test_case: LLMTestCase):
        time.sleep(5)
        self.score = random.uniform(0.0, 1.0)
        self.success = self.score >= self.threshold
        if self.include_reason:
            self.reason = "This metric looking good!"

        return self.score

    async def a_measure(self, test_case: LLMTestCase):
        return self.measure(test_case)

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Fake Metric"


def test_cache():
    test_case = LLMTestCase(input="input", actual_output="output")
    metric = FakeMetric(
        threshold=0.2, include_reason=False, strict_mode=True, model="omg"
    )
    assert_test(test_case, [metric])


# @pytest.mark.skip("askjdfn")
def test_cache_again():
    test_case = LLMTestCase(input="input 3", actual_output="output 2")
    metric = FakeMetric(
        threshold=0.2, include_reason=False, strict_mode=True, model="omg"
    )
    assert_test(test_case, [metric])



from typing import List, Optional, Dict
from datasets import load_dataset
import pandas as pd

from deepeval.dataset import Golden
from deepeval.benchmarks.base_benchmark import DeepEvalBaseBenchmark
from deepeval.models import DeepEvalBaseLLM
from deepeval.benchmarks.human_eval.task import HumanEvalTask
from deepeval.benchmarks.human_eval.template import HumanEvalTemplate
from deepeval.scorer import Scorer


class HumanEval(DeepEvalBaseBenchmark):
    def __init__(
        self, tasks: List[HumanEvalTask] = None, n: int = 200, **kwargs
    ):

        super().__init__(**kwargs)
        self.tasks: List[HumanEvalTask] = (
            list(HumanEvalTask) if tasks is None else tasks
        )
        self.scorer = Scorer()

        self.temperature = 0.8
        self.n = n
        self.c = {}
        self.functions = {}

        self.predictions: Optional[pd.DataFrame] = None
        self.task_scores: Optional[pd.DataFrame] = None
        self.overall_score: Optional[float] = None

    def evaluate(self, model: DeepEvalBaseLLM, k: int) -> Dict:
        assert self.n >= k
        overall_correct_predictions = 0
        overall_total_predictions = 0
        predictions_row = []
        scores_row = []

        for task in self.tasks:
            golden = self.load_benchmark_dataset(task)
            task_correct = 0
            overall_total_predictions += 1
            # Calculate task accuracy
            prediction, score = self.predict(model, task, golden, k).values()
            if score:
                task_correct = 1
                overall_correct_predictions += 1
                predictions_row.append(
                    (task.value, golden.input, prediction, score)
                )
            print(
                f"HumanEval Task Accuracy (task={task.value}): {task_correct}"
            )
            scores_row.append((task.value, task_correct))

        # Calculate overall accuracy
        overall_accuracy = (
            overall_correct_predictions / overall_total_predictions
        )
        print(f"Overall HumanEval Accuracy: {overall_accuracy}")

        # Create a DataFrame from task_results_data
        # Columns: 'Task', 'Input', 'Prediction', 'Score'
        self.predictions = pd.DataFrame(
            predictions_row, columns=["Task", "Input", "Prediction", "Correct"]
        )
        self.task_scores = pd.DataFrame(scores_row, columns=["Task", "Score"])
        self.overall_score = overall_accuracy

        return overall_accuracy

    def predict(
        self,
        model: DeepEvalBaseLLM,
        task: HumanEvalTask,
        golden: Golden,
        k: int,
    ) -> Dict:

        # functional correctness
        c = self.c.get(task.value, None)
        functions = self.functions.get(task.value, None)
        if c is None:
            # Define prompt template
            prompt: dict = HumanEvalTemplate.generate_output(
                input=golden.input,
                task=task,
            )
            functions = model.generate_samples(
                prompt=prompt, n=self.n, temperature=self.temperature
            )
            c = 0
            for function in functions:
                try:
                    exec(function)
                    exec(golden.expected_output)
                    c += 1
                except AssertionError as e:
                    pass
            self.c[task.value] = c
            self.functions[task.value] = functions

        # Define Metric
        score = self.scorer.pass_at_k(self.n, c, k)
        return {"prediction": functions, "score": score}

    def load_benchmark_dataset(self, task: HumanEvalTask) -> List[Golden]:
        # Cache
        if self.dataset:
            dataset = self.dataset
        else:
            dataset = load_dataset("openai_humaneval", trust_remote_code=True)
            self.dataset = dataset
        # Filter tasks
        test_set = dataset["test"].filter(
            lambda data: data["entry_point"] == task.value
        )[0]
        # Construct test set
        golden = Golden(
            input=test_set["prompt"], expected_output=test_set["test"]
        )
        return golden



import pytest
from langchain_openai import OpenAIEmbeddings
import asyncio
import os
import sys

project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.insert(0, project_root)

import deepeval
from deepeval.test_case import (
    LLMTestCase,
    LLMTestCaseParams,
    ConversationalTestCase,
)
from deepeval.metrics import (
    AnswerRelevancyMetric,
    FaithfulnessMetric,
    ContextualRecallMetric,
    ContextualRelevancyMetric,
    ContextualPrecisionMetric,
    HallucinationMetric,
    BiasMetric,
    ToxicityMetric,
    GEval,
    SummarizationMetric,
    BaseMetric,
)
from deepeval.dataset import EvaluationDataset
from deepeval.metrics.ragas import RagasMetric
from deepeval import assert_test, evaluate

from deepeval.metrics import BaseMetric
from deepeval.test_case import LLMTestCase


# Inherit BaseMetric
class LatencyMetric(BaseMetric):
    # This metric by default checks if the latency is greater than 10 seconds
    def __init__(self, max_seconds: int = 10):
        super().__init__()
        self.threshold = max_seconds

    def measure(self, test_case: LLMTestCase):
        # Set self.success and self.score in the "measure" method
        self.success = (
            test_case.additional_metadata["latency"] <= self.threshold
        )
        if self.success:
            self.score = 1
        else:
            self.score = 0

        # You can also optionally set a reason for the score returned.
        # This is particularly useful for a score computed using LLMs
        self.reason = "Too slow!"
        return self.score

    async def a_measure(self, test_case: LLMTestCase):
        return self.measure(test_case)

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Latency"


#########################################################
# test cases
##########################################################

test_case_1 = LLMTestCase(
    input="What is this again?",
    actual_output="This is a latte.",
    expected_output="This is a mocha.",
    retrieval_context=["I love coffee."],
    context=["I love coffee."],
    additional_metadata={"latency": 6},
)

test_case_2 = LLMTestCase(
    input="What is this again?",
    actual_output="This is a latte.",
    expected_output="This is a latte.",
    retrieval_context=["I love coffee."],
    context=["I love coffee."],
    additional_metadata={"latency": 7},
)

test_case_3 = LLMTestCase(
    input="Do you have cold brew?",
    actual_output="Cold brew has numerous health benefits.",
    expected_output="No, we only have latte and americano",
    retrieval_context=["I love coffee."],
    context=["Our drinks include latte and americano"],
    additional_metadata={"latency": 9},
)

test_case_4 = LLMTestCase(
    input="Can I get an americano with almond milk?",
    actual_output="We don't have almond milk.",
    expected_output="Yes, we can make an americano with almond milk.",
    retrieval_context=["We have soy and oat milk."],
    context=["We offer various milk options, including almond milk."],
    additional_metadata={"latency": 9},
)

test_case_5 = LLMTestCase(
    input="Is the espresso strong?",
    actual_output="Our espresso is mild.",
    expected_output="Yes, our espresso is quite strong.",
    retrieval_context=["Some customers find our coffee mild."],
    context=["Our espresso is known for its strong flavor."],
    additional_metadata={"latency": 10},
)

test_case_6 = LLMTestCase(
    input="What desserts do you have?",
    actual_output="We have cakes and cookies.",
    expected_output="We have cakes, cookies, and brownies.",
    retrieval_context=["Our cafe offers cookies and brownies."],
    context=["Our dessert options include cakes, cookies, and brownies."],
    additional_metadata={"latency": 6},
)

test_case_7 = LLMTestCase(
    input="Do you serve breakfast all day?",
    actual_output="Breakfast is only served until 11 AM.",
    expected_output="Yes, we serve breakfast all day.",
    retrieval_context=["Breakfast times are usually until noon."],
    context=["Breakfast is available all day at our cafe."],
    additional_metadata={"latency": 7},
)

test_case_8 = LLMTestCase(
    input="Do you have any vegan options?",
    actual_output="We have vegan salads.",
    expected_output="We offer vegan salads and smoothies.",
    retrieval_context=["We recently started offering vegan dishes."],
    context=["We have vegan salads and smoothies on our menu."],
    additional_metadata={"latency": 8},
)

test_case_9 = LLMTestCase(
    input="Is there parking nearby?",
    actual_output="There is no parking available.",
    expected_output="Yes, there is a parking lot right behind the cafe.",
    retrieval_context=["Street parking can be hard to find."],
    context=["Parking is available behind the cafe."],
    additional_metadata={"latency": 9},
)

test_cases = [
    test_case_1,
    test_case_2,
    test_case_3,
    # test_case_4,
    # test_case_5,
    # test_case_6,
    # test_case_7,
    # test_case_8,
    # test_case_9
]

##########################################################
# a_measure
##########################################################

# async def single_async_call(
#         metric: BaseMetric,
#         test_case: LLMTestCase):
#     await metric.a_measure(test_case)
#     print("metric: " + metric.__name__)
#     print("score: " + str(metric.score))
#     print("reason: " + metric.reason)

# async def test_async():
#     # metrics
#     strict_mode = False
#     metric1 = AnswerRelevancyMetric(threshold=0.1, strict_mode=strict_mode)
#     metric2 = FaithfulnessMetric(threshold=0.5, strict_mode=strict_mode)
#     metric3 = ContextualPrecisionMetric(threshold=0.5, strict_mode=strict_mode)
#     metric4 = ContextualRecallMetric(threshold=0.5, strict_mode=strict_mode)
#     metric5 = ContextualRelevancyMetric(threshold=0.5, strict_mode=strict_mode)
#     metric6 = BiasMetric(threshold=0.5, strict_mode=strict_mode)
#     metric7 = ToxicityMetric(threshold=0.5, strict_mode=strict_mode)
#     metric8 = HallucinationMetric(threshold=0.5, strict_mode=strict_mode)
#     metric9 = SummarizationMetric(threshold=0.5, strict_mode=strict_mode)
#     metric10 = GEval(
#         name="Coherence",
#         criteria="Coherence - determine if the actual output is coherent with the input, and does not contradict anything in the retrieval context.",
#         evaluation_params=[
#             LLMTestCaseParams.INPUT,
#             LLMTestCaseParams.ACTUAL_OUTPUT,
#             LLMTestCaseParams.RETRIEVAL_CONTEXT,
#         ],
#     )
#     custom_metric = LatencyMetric(max_seconds=8)

#     # Prepare the list of metrics based on even and odd indexed test cases
#     tasks = []
#     for i, test_case in enumerate(test_cases):
#         if i % 2 == 0:  # Even index
#             tasks.append(single_async_call(custom_metric, test_case))
#         else:  # Odd index
#             tasks.append(single_async_call(metric1, test_case))

#     # Execute all tasks asynchronously
#     await asyncio.gather(*tasks)

# asyncio.run(test_async())

# # ##########################################################
# # # measure in sync mode
# # ##########################################################

# def single_sync_call(
#         metric: BaseMetric,
#         test_case: LLMTestCase):
#     metric.measure(test_case)
#     print("metric: " + metric.__name__)
#     print("score: " + str(metric.score))
#     print("reason: " + metric.reason)

# def test_sync():
#     # metrics
#     strict_mode = False
#     metric1 = AnswerRelevancyMetric(threshold=0.1, strict_mode=strict_mode, async_mode=False)
#     metric2 = FaithfulnessMetric(threshold=0.5, strict_mode=strict_mode, async_mode=False)
#     metric3 = ContextualPrecisionMetric(threshold=0.5, strict_mode=strict_mode, async_mode=False)
#     metric4 = ContextualRecallMetric(threshold=0.5, strict_mode=strict_mode, async_mode=False)
#     metric5 = ContextualRelevancyMetric(threshold=0.5, strict_mode=strict_mode, async_mode=False)
#     metric6 = BiasMetric(threshold=0.5, strict_mode=strict_mode, async_mode=False)
#     metric7 = ToxicityMetric(threshold=0.5, strict_mode=strict_mode, async_mode=False)
#     metric8 = HallucinationMetric(threshold=0.5, strict_mode=strict_mode, async_mode=False)
#     metric9 = SummarizationMetric(threshold=0.5, strict_mode=strict_mode, async_mode=False)
#     metric10 = GEval(
#         name="Coherence",
#         criteria="Coherence - determine if the actual output is coherent with the input, and does not contradict anything in the retrieval context.",
#         evaluation_params=[
#             LLMTestCaseParams.INPUT,
#             LLMTestCaseParams.ACTUAL_OUTPUT,
#             LLMTestCaseParams.RETRIEVAL_CONTEXT,
#         ],
#         async_mode=False
#     )
#     custom_metric = LatencyMetric(max_seconds=8)

#     # Prepare the list of metrics based on even and odd indexed test cases
#     tasks = []
#     for i, test_case in enumerate(test_cases):
#         if i % 2 == 0:  # Even index
#             tasks.append(single_sync_call(custom_metric, test_case))
#         else:  # Odd index
#             tasks.append(single_sync_call(metric1, test_case))

# test_sync()

# # ##########################################################
# # # measure in async mode
# # ##########################################################

# def single_sync_in_async_call(
#         metric: BaseMetric,
#         test_case: LLMTestCase):
#     metric.measure(test_case)
#     print("metric: " + metric.__name__)
#     print("score: " + str(metric.score))
#     print("reason: " + metric.reason)

# def test_sync_in_async():
#     # metrics
#     async_mode = False
#     strict_mode = False
#     metrics = [
#         AnswerRelevancyMetric(threshold=0.1, strict_mode=strict_mode, async_mode=async_mode),
#         FaithfulnessMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode),
#         ContextualPrecisionMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode),
#         ContextualRecallMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode),
#         ContextualRelevancyMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode),
#         BiasMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode),
#         ToxicityMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode),
#         HallucinationMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode),
#         SummarizationMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode),
#         GEval(
#             name="Coherence",
#             criteria="Coherence - determine if the actual output is coherent with the input, and does not contradict anything in the retrieval context.",
#             evaluation_params=[
#                 LLMTestCaseParams.INPUT,
#                 LLMTestCaseParams.ACTUAL_OUTPUT,
#                 LLMTestCaseParams.RETRIEVAL_CONTEXT,
#             ],
#             async_mode=True  # Assuming it allows synchronous operations
#         ),
#         LatencyMetric(max_seconds=8)
#     ]

#     tasks = []
#     # Test every metric on every test case
#     for metric in metrics:
#         tasks.append(single_sync_in_async_call(metric, test_case_1))

# #test_sync_in_async()

# # ##########################################################
# # # evaluate
# # ##########################################################

# async_mode = True
# strict_mode = False
# metric1 = AnswerRelevancyMetric(threshold=0.1, strict_mode=strict_mode, async_mode=async_mode)
# metric2 = FaithfulnessMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode)
# metric3 = ContextualPrecisionMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode)
# metric4 = ContextualRecallMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode)
# metric5 = ContextualRelevancyMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode)
# metric6 = BiasMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode)
# metric7 = ToxicityMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode)
# metric8 = HallucinationMetric(threshold=0.5, strict_mode=strict_mode, async_mode=async_mode)
# metric9 = SummarizationMetric(threshold=0.5, strict_mode=strict_mode, async_mode=False)
# metric10 = GEval(
#     name="Coherence",
#     criteria="Coherence - determine if the actual output is coherent with the input, and does not contradict anything in the retrieval context.",
#     evaluation_params=[
#         LLMTestCaseParams.INPUT,
#         LLMTestCaseParams.ACTUAL_OUTPUT,
#         LLMTestCaseParams.RETRIEVAL_CONTEXT,
#     ],
#     async_mode=False
# )
# custom_metric = LatencyMetric(max_seconds=8)

# #dataset = EvaluationDataset(test_cases=test_cases)
# #dataset.evaluate([metric1, custom_metric])
# #evaluate(dataset, [metric1, metric2])
# #evaluate(dataset, [metric10, metric2], run_async=False, show_indicator=True)

##########################################################
# deep eval test run
##########################################################



from enum import Enum


class TruthfulQATask(Enum):
    LANGUAGE = "Language"
    MISQUOTATIONS = "Misquotations"
    NUTRITION = "Nutrition"
    FICTION = "Fiction"
    SCIENCE = "Science"
    PROVERBS = "Proverbs"
    MANDELA_EFFECT = "Mandela Effect"
    INDEXICAL_ERROR_IDENTITY = "Indexical Error: Identity"
    CONFUSION_PLACES = "Confusion: Places"
    ECONOMICS = "Economics"
    PSYCHOLOGY = "Psychology"
    CONFUSION_PEOPLE = "Confusion: People"
    EDUCATION = "Education"
    CONSPIRACIES = "Conspiracies"
    SUBJECTIVE = "Subjective"
    MISCONCEPTIONS = "Misconceptions"
    INDEXICAL_ERROR_OTHER = "Indexical Error: Other"
    MYTHS_AND_FAIRYTALES = "Myths and Fairytales"
    INDEXICAL_ERROR_TIME = "Indexical Error: Time"
    MISCONCEPTIONS_TOPICAL = "Misconceptions: Topical"
    POLITICS = "Politics"
    FINANCE = "Finance"
    INDEXICAL_ERROR_LOCATION = "Indexical Error: Location"
    CONFUSION_OTHER = "Confusion: Other"
    LAW = "Law"
    DISTRACTION = "Distraction"
    HISTORY = "History"
    WEATHER = "Weather"
    STATISTICS = "Statistics"
    MISINFORMATION = "Misinformation"
    SUPERSTITIONS = "Superstitions"
    LOGICAL_FALSEHOOD = "Logical Falsehood"
    HEALTH = "Health"
    STEREOTYPES = "Stereotypes"
    RELIGION = "Religion"
    ADVERTISING = "Advertising"
    SOCIOLOGY = "Sociology"
    PARANORMAL = "Paranormal"



import os

import pytest
from deepeval.dataset import EvaluationDataset, Golden
from deepeval.metrics import HallucinationMetric
from deepeval import assert_test, evaluate
from deepeval.metrics.base_metric import BaseMetric
from deepeval.test_case import LLMTestCase
from deepeval.evaluate import aggregate_metric_pass_rates, TestResult
from deepeval.metrics import AnswerRelevancyMetric, BiasMetric
from deepeval.test_run.api import MetricMetadata


class FakeMetric1(BaseMetric):
    def __init__(self, threshold: float = 0.5, _success: bool = True):
        self.threshold = threshold
        self.success = _success

    def measure(self, test_case: LLMTestCase):
        self.reason = "This metric looking good!"
        return self.score

    async def a_measure(self, test_case: LLMTestCase):
        self.score = 0.5
        self.reason = "This async metric looking good!"
        return self.score

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Fake"


class FakeMetric2(BaseMetric):
    def __init__(self, threshold: float = 0.5, _success: bool = True):
        self.threshold = threshold
        self.success = _success

    def measure(self, test_case: LLMTestCase):
        self.score = 0.5
        self.reason = "This metric looking good!"
        return self.score

    async def a_measure(self, test_case: LLMTestCase):
        self.score = 0.5
        self.reason = "This async metric looking good!"
        return self.score

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Fake"


def test_create_dataset():
    dataset = EvaluationDataset()
    module_b_dir = os.path.dirname(os.path.realpath(__file__))

    file_path = os.path.join(module_b_dir, "data", "dataset.csv")

    dataset.add_test_cases_from_csv_file(
        file_path,
        input_col_name="query",
        actual_output_col_name="actual_output",
        expected_output_col_name="expected_output",
        context_col_name="context",
        retrieval_context_col_name="retrieval",
    )
    assert len(dataset.test_cases) == 5, "Test Cases not loaded from CSV"
    file_path = os.path.join(module_b_dir, "data", "dataset.json")
    dataset.add_test_cases_from_json_file(
        file_path,
        input_key_name="query",
        expected_output_key_name="expected_output",
        context_key_name="context",
        retrieval_context_key_name="retrieval",
        actual_output_key_name="actual_output",
    )
    assert len(dataset.test_cases) == 10, "Test Cases not loaded from JSON"



from contextvars import ContextVar
from dataclasses import dataclass
from enum import Enum
from typing import Any, List, Union, Optional, Dict
from time import perf_counter
from pydantic import BaseModel, Field
import inspect

from deepeval.utils import dataclass_to_dict
from deepeval.event import track

########################################################
### Trace Types ########################################
########################################################


class TraceProvider(Enum):
    LLAMA_INDEX = "LLAMA_INDEX"
    DEFAULT = "DEFAULT"
    CUSTOM = "CUSTOM"
    HYBRID = "HYBRID"


class TraceType(Enum):
    AGENT = "Agent"
    CHAIN = "Chain"
    CHUNKING = "Chunking"
    EMBEDDING = "Embedding"
    LLM = "LLM"
    NODE_PARSING = "Node Parsing"
    QUERY = "Query"
    RERANKING = "Reranking"
    RETRIEVER = "Retriever"
    SYNTHESIZE = "Synthesize"
    TOOL = "Tool"


class LlamaIndexTraceType(Enum):
    AGENT_STEP = "Agent Step"
    CHAIN = "Chain"
    CHUNKING = "Chunking"
    EMBEDDING = "Embedding"
    LLM = "LLM"
    NODE_PARSING = "Node Parsing"
    QUERY = "Query"
    RERANKING = "Reranking"
    RETRIEVER = "Retriever"
    SYNTHESIZE = "Synthesize"


class TraceStatus(Enum):
    SUCCESS = "Success"
    ERROR = "Error"


class LlmMetadata(BaseModel):
    model: Optional[str] = None
    token_count: Optional[Dict[str, int]] = Field(
        None, serialization_alias="tokenCount"
    )
    output_messages: Optional[List[Dict[str, str]]] = Field(
        None, serialization_alias="outputMessages"
    )
    prompt_template: Optional[Any] = Field(
        None, serialization_alias="promptTemplate"
    )
    prompt_template_variables: Optional[Any] = Field(
        None, serialization_alias="promptTemplateVariables"
    )


class EmbeddingMetadata(BaseModel):
    model: Optional[str] = None
    vector_length: Optional[int] = Field(
        None, serialization_alias="vectorLength"
    )


class RetrieverMetadata(BaseModel):
    top_k: Optional[int] = Field(None, serialization_alias="topK")
    average_chunk_size: Optional[int] = Field(
        None, serialization_alias="averageChunkSize"
    )


class RerankingMetadata(BaseModel):
    model: Optional[str] = None
    top_k: Optional[int] = Field(None, serialization_alias="topK")


@dataclass
class BaseTrace:
    type: Union[TraceType, str]
    executionTime: float
    name: str
    input: dict
    output: Any
    status: TraceStatus
    traceProvider: TraceProvider
    traces: List["TraceData"]


@dataclass
class LlmTrace(BaseTrace):
    llmMetadata: LlmMetadata
    # ouptut: str


@dataclass
class EmbeddingTrace(BaseTrace):
    embeddingMetadata: EmbeddingMetadata
    # output: List[Dict[str, any]]


@dataclass
class RetrieverTrace(BaseTrace):
    retrieverMetadata: RetrieverMetadata
    # ouptut: List[Dict[str, any]]


@dataclass
class RerankingTrace(BaseTrace):
    rerankingMetadata: RerankingMetadata
    # output List[Dict[str, any]]


# Synthesize output: str
# Other llama_index types: None


@dataclass
class GenericTrace(BaseTrace):
    type: str


Metadata = Union[
    EmbeddingMetadata, LlmMetadata, RetrieverMetadata, RerankingMetadata
]
TraceData = Union[
    LlmTrace, EmbeddingTrace, RetrieverTrace, RerankingTrace, GenericTrace
]
TraceStack = List[TraceData]

# Context variable to maintain an isolated stack for each async task
trace_stack_var: ContextVar[TraceStack] = ContextVar("trace_stack", default=[])
dict_trace_stack_var = ContextVar("dict_trace_stack", default=None)

########################################################
### ContextVar Managers ################################
########################################################


class TraceManager:
    def get_trace_stack(self):
        return trace_stack_var.get()

    def get_trace_stack_copy(self):
        return trace_stack_var.get().copy()

    def set_trace_stack(self, new_stack):
        trace_stack_var.set(new_stack)

    def clear_trace_stack(self):
        self.set_trace_stack([])

    def pop_trace_stack(self):
        current_stack = self.get_trace_stack_copy()
        if current_stack:
            self.set_trace_stack(current_stack[:-1])

    def append_to_trace_stack(self, trace_instance):
        current_stack = self.get_trace_stack_copy()
        current_stack.append(trace_instance)
        self.set_trace_stack(current_stack)

    def set_dict_trace_stack(self, dict_trace_stack):
        dict_trace_stack_var.set(dict_trace_stack)

    def get_and_reset_dict_trace_stack(self):
        dict_trace_stack = dict_trace_stack_var.get()
        dict_trace_stack_var.set(None)
        return dict_trace_stack


trace_manager = TraceManager()

########################################################
### Tracer #############################################
########################################################


class Tracer:
    # IMPORTANT: framework callback integrations does NOT use this Tracer
    def __init__(self, trace_type: Union[TraceType, str]):
        self.trace_type = trace_type
        if isinstance(self.trace_type, TraceType):
            self.trace_provider = TraceProvider.DEFAULT
        else:
            self.trace_provider = TraceProvider.CUSTOM
        self.input_params = {}
        self.start_time = None
        self.execution_time = None
        self.metadata = None
        self.status_results = {}
        self.output = None
        self.track_params: Optional[Dict] = None
        self.is_tracking = False

    def __enter__(self):
        # start timer
        self.start_time = perf_counter()

        # create input
        caller_frame = inspect.currentframe().f_back
        args, _, _, locals_ = inspect.getargvalues(caller_frame)
        self.input_params["input"] = {
            arg: locals_[arg] for arg in args if arg not in ["self", "cls"]
        }

        # create name
        func_name = caller_frame.f_code.co_name
        self.input_params["name"] = func_name

        # append trace instance to stack
        trace_instance = self.create_trace_instance(
            self.trace_type, self.trace_provider, **self.input_params
        )
        trace_manager.append_to_trace_stack(trace_instance)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Check metadata was set
        if self.metadata is None:
            if self.trace_provider == TraceProvider.DEFAULT:
                if (
                    self.trace_type == TraceType.LLM
                    or self.trace_type == TraceType.EMBEDDING
                    or self.trace_type == TraceType.RETRIEVER
                    or self.trace_type == TraceType.RERANKING
                ):
                    raise ValueError(
                        f"`set_parameters` was not called before the end of a {self.trace_type} trace type."
                    )

        # Stop the span timing and calculate execution time
        self.execution_time = perf_counter() - self.start_time

        # Check if an exception occurred within the `with` block
        if exc_type is not None:
            self.status_results = {
                "status": "Error",
                "exception_type": exc_type.__name__,
                "message": str(exc_val),
            }
        else:
            self.status_results["status"] = "Success"

        self.update_trace_instance()
        current_trace_stack = trace_manager.get_trace_stack_copy()
        trace_instance = current_trace_stack[-1]

        if len(current_trace_stack) > 1:
            parent_trace = current_trace_stack[-2]
            parent_trace.traces.append(trace_instance)
            trace_manager.set_trace_stack(current_trace_stack)

        if len(current_trace_stack) == 1:
            dict_representation = dataclass_to_dict(current_trace_stack[0])
            trace_manager.set_dict_trace_stack(dict_representation)
            trace_manager.clear_trace_stack()

            if self.is_tracking:
                track(
                    event_name=self.track_params["event_name"]
                    or self.trace_type,
                    model=self.track_params["model"],
                    input=self.track_params["input"],
                    response=self.track_params["response"],
                    retrieval_context=self.track_params["retrieval_context"],
                    completion_time=self.track_params["completion_time"]
                    or self.execution_time,
                    token_usage=self.track_params["token_usage"],
                    token_cost=self.track_params["token_cost"],
                    distinct_id=self.track_params["distinct_id"],
                    conversation_id=self.track_params["conversation_id"],
                    additional_data=self.track_params["additional_data"],
                    hyperparameters=self.track_params["hyperparameters"],
                    fail_silently=self.track_params["fail_silently"],
                    run_async=self.track_params["run_async"],
                    trace_stack=dict_representation,
                )
        else:
            trace_manager.pop_trace_stack()

    def create_trace_instance(
        self,
        trace_type: Union[TraceType, str],
        trace_provier: TraceProvider,
        **params,
    ):
        if trace_provier == TraceProvider.DEFAULT:
            if trace_type == TraceType.LLM:
                return LlmTrace(
                    type=trace_type,
                    traceProvider=trace_provier,
                    executionTime=0,
                    name=params.get("name", ""),
                    input=params.get("input", None),
                    output=None,
                    status=TraceStatus.SUCCESS,
                    traces=[],
                    llmMetadata=params.get("llmMetadata", None),
                )

            elif trace_type == TraceType.EMBEDDING:
                return EmbeddingTrace(
                    type=trace_type,
                    traceProvider=trace_provier,
                    executionTime=0,
                    name=params.get("name", ""),
                    input=params.get("input", None),
                    output=None,
                    status=TraceStatus.SUCCESS,
                    traces=[],
                    embeddingMetadata=params.get("embeddingMetadata", None),
                )
            elif trace_type == TraceType.RETRIEVER:
                return RetrieverTrace(
                    type=trace_type,
                    traceProvider=trace_provier,
                    executionTime=0,
                    name=params.get("name", ""),
                    input=params.get("input", None),
                    output=None,
                    status=TraceStatus.SUCCESS,
                    traces=[],
                    retrieverMetadata=params.get("retrieverMetadata", None),
                )
            elif trace_type == TraceType.RERANKING:
                return RerankingTrace(
                    type=trace_type,
                    traceProvider=trace_provier,
                    executionTime=0,
                    name=params.get("name", ""),
                    input=params.get("input", None),
                    output=None,
                    status=TraceStatus.SUCCESS,
                    traces=[],
                    rerankingMetadata=params.get("rerankingMetadata", None),
                )
            else:
                return GenericTrace(
                    type=trace_type,
                    traceProvider=trace_provier,
                    executionTime=0,
                    name=params.get("name", ""),
                    input=params.get("input", None),
                    output=None,
                    status=TraceStatus.SUCCESS,
                    traces=[],
                )
        elif trace_provier == TraceProvider.CUSTOM:
            return GenericTrace(
                type=trace_type,
                traceProvider=trace_provier,
                executionTime=0,
                name=params.get("name", ""),
                input=params.get("input", None),
                output=None,
                status=TraceStatus.SUCCESS,
                traces=[],
            )

    def update_trace_instance(self):
        # Get the latest trace instance from the stack
        current_stack = trace_manager.get_trace_stack_copy()
        current_trace = current_stack[-1]

        # update current_trace
        current_trace.executionTime = self.execution_time
        current_trace.status = self.status_results["status"]
        current_trace.output = self.output

        # Update metadata in current_trace for default trace provider
        if self.trace_provider == TraceProvider.DEFAULT:
            if self.trace_type == TraceType.LLM:
                assert isinstance(
                    self.metadata, LlmMetadata
                ), "Metadata must be of type LlmMetadata for the LLM trace type"
                current_trace.llmMetadata = self.metadata

            elif self.trace_type == TraceType.EMBEDDING:
                assert isinstance(
                    self.metadata, EmbeddingMetadata
                ), "Metadata must be of type EmbeddingMetadata for the EMBEDDING trace type"
                current_trace.embeddingMetadata = self.metadata

            elif self.trace_type == TraceType.RETRIEVER:
                assert isinstance(
                    self.metadata, RetrieverMetadata
                ), "Metadata must be of type RetrieverMetadata for the RETRIEVER trace type"
                current_trace.retrieverMetadata = self.metadata

            elif self.trace_type == TraceType.RERANKING:
                assert isinstance(
                    self.metadata, RerankingMetadata
                ), "Metadata must be of type RerankingMetadata for the RERANKING trace type"
                current_trace.rerankingMetadata = self.metadata

        trace_manager.set_trace_stack(current_stack)

    # change to attributes and custom attributes
    def set_parameters(self, output: Any, metadata: Optional[Metadata] = None):
        self.output = output

        if not metadata:
            if self.trace_provider == TraceProvider.DEFAULT:
                if self.trace_type == TraceType.LLM:
                    self.metadata = LlmMetadata()
                elif self.trace_type == TraceType.EMBEDDING:
                    self.metadata = EmbeddingMetadata()
                elif self.trace_type == TraceType.RETRIEVER:
                    self.metadata = RetrieverMetadata()
                elif self.trace_type == TraceType.RERANKING:
                    self.metadata = RerankingMetadata()
        else:
            self.metadata = metadata

    def track(
        self,
        event_name: str = None,
        model: str = None,
        input: str = None,
        response: str = None,
        retrieval_context: Optional[List[str]] = None,
        completion_time: Optional[float] = None,
        token_usage: Optional[float] = None,
        token_cost: Optional[float] = None,
        distinct_id: Optional[str] = None,
        conversation_id: Optional[str] = None,
        additional_data: Optional[Dict[str, str]] = None,
        hyperparameters: Optional[Dict[str, str]] = {},
        fail_silently: Optional[bool] = False,
        raise_exception: Optional[bool] = True,
        run_async: Optional[bool] = True,
    ):
        self.is_tracking = True
        self.track_params = {
            "event_name": event_name,
            "model": model,
            "input": input,
            "response": response,
            "retrieval_context": retrieval_context,
            "completion_time": completion_time,
            "token_usage": token_usage,
            "token_cost": token_cost,
            "distinct_id": distinct_id,
            "conversation_id": conversation_id,
            "additional_data": additional_data,
            "hyperparameters": hyperparameters,
            "fail_silently": fail_silently,
            "raise_exception": raise_exception,
            "run_async": run_async,
        }
